"""Data access object for timetable"""
import os
from pathlib import Path

from loguru import logger
import joblib

from pyraptor.model.structures import Timetable
from pyraptor.util import mkdir_if_not_exists


def read_timetable(input_folder: str) -> Timetable:
    """
    Read the timetable data from the cache directory
    """

    def load_joblib(name):
        logger.debug(f"Loading '{name}'")
        with open(Path(input_folder, f"{name}.pcl"), "rb") as handle:
            return joblib.load(handle)

    if not os.path.exists(input_folder):
        raise IOError(
            "PyRaptor timetable not found. Run `python pyraptor/gtfs/timetable.py`"
            " first to create timetable from GTFS files."
        )

    logger.debug("Using cached datastructures")

    timetable: Timetable = load_joblib("timetable")

    return timetable


def write_timetable(output_folder: str, timetable: Timetable) -> None:
    """
    Write the timetable to output directory
    """

    def write_joblib(state, name):
        with open(Path(output_folder, f"{name}.pcl"), "wb") as handle:
            joblib.dump(state, handle)

    logger.info("Write PyRaptor timetable to output directory")

    mkdir_if_not_exists(output_folder)
    write_joblib(timetable, "timetable")



"""Shared functionality between RAPTOR algorithms"""
from typing import List

from loguru import logger

from pyraptor.model.structures import Journey
from pyraptor.util import sec2str


def print_journeys(journeys: List[Journey], dep_secs=None):
    """Print list of journeys"""
    for jrny in journeys:
        print_journey(jrny, dep_secs)


def print_journey(journey: Journey, dep_secs=None):
    """Print the given journey to logger info"""

    logger.info("Journey:")

    if len(journey) == 0:
        logger.info("No journey available")
        return

    # Print all legs in journey
    for leg in journey:
        # Start and end stop of leg and trip
        msg = (
            str(sec2str(leg.dep))
            + " "
            + leg.from_stop.station.name.ljust(20)
            + "(p. "
            + str(leg.from_stop.platform_code).rjust(3)
            + ") TO "
            + str(sec2str(leg.arr))
            + " "
            + leg.to_stop.station.name.ljust(20)
            + "(p. "
            + str(leg.to_stop.platform_code).rjust(3)
            + ") WITH "
            + str(leg.trip.hint)
        )
        logger.info(msg)

    logger.info(f"Fare: €{journey.fare()}")

    msg = f"Duration: {sec2str(journey.arr() - journey.dep())}"
    if dep_secs:
        msg += " ({} from request time {})".format(
            sec2str(journey.arr() - dep_secs),
            sec2str(dep_secs),
        )
    logger.info(msg)
    logger.info("")



"""Utility functions"""
import os
import numpy as np


TRANSFER_COST = 2 * 60  # Default transfer time is 2 minutes
LARGE_NUMBER = 999999
TRANSFER_TRIP = None


def mkdir_if_not_exists(name: str) -> None:
    """Create directory if not exists"""
    if not os.path.exists(name):
        os.makedirs(name)


def str2sec(time_str: str) -> int:
    """
    Convert hh:mm:ss to seconds since midnight
    :param time_str: String in format hh:mm:ss
    """
    split_time = time_str.strip().split(":")
    if len(split_time) == 3:
        # Has seconds
        hours, minutes, seconds = split_time
        return int(hours) * 3600 + int(minutes) * 60 + int(seconds)
    minutes, seconds = split_time
    return int(minutes) * 60 + int(seconds)


def sec2str(scnds: int, show_sec: bool = False) -> str:
    """
    Convert hh:mm:ss to seconds since midnight

    :param show_sec: only show :ss if True
    :param scnds: Seconds to translate to hh:mm:ss
    """
    scnds = np.round(scnds)
    hours = int(scnds / 3600)
    minutes = int((scnds % 3600) / 60)
    seconds = int(scnds % 60)
    return (
        "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)
        if show_sec
        else "{:02d}:{:02d}".format(hours, minutes)
    )



"""Run query with RAPTOR algorithm"""
import argparse
from typing import List
from copy import copy

from loguru import logger

from pyraptor.dao.timetable import read_timetable
from pyraptor.model.base import print_journeys
from pyraptor.model.structures import Timetable, Journey
from pyraptor.model.mcraptor import (
    McRaptorAlgorithm,
    reconstruct_journeys,
    best_legs_to_destination_station,
)
from pyraptor.util import str2sec


def parse_arguments():
    """Parse arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="data/output",
        help="Input directory",
    )
    parser.add_argument(
        "-or",
        "--origin",
        type=str,
        default="Hertogenbosch ('s)",
        help="Origin station of the journey",
    )
    parser.add_argument(
        "-d",
        "--destination",
        type=str,
        default="Rotterdam Centraal",
        help="Destination station of the journey",
    )
    parser.add_argument(
        "-t", "--time", type=str, default="08:35:00", help="Departure time (hh:mm:ss)"
    )
    parser.add_argument(
        "-r",
        "--rounds",
        type=int,
        default=4,
        help="Number of rounds to execute the RAPTOR algorithm",
    )
    arguments = parser.parse_args()
    return arguments


def main(
    input_folder,
    origin_station,
    destination_station,
    departure_time,
    rounds,
):
    """Run RAPTOR algorithm"""

    logger.debug("Input directory     : {}", input_folder)
    logger.debug("Origin station      : {}", origin_station)
    logger.debug("Destination station : {}", destination_station)
    logger.debug("Departure time      : {}", departure_time)
    logger.debug("Rounds              : {}", str(rounds))

    timetable = read_timetable(input_folder)

    logger.info(f"Calculating network from : {origin_station}")

    # Departure time seconds
    dep_secs = str2sec(departure_time)
    logger.debug("Departure time (s.)  : " + str(dep_secs))

    # Find route between two stations
    journeys = run_mcraptor(
        timetable,
        origin_station,
        destination_station,
        dep_secs,
        rounds,
    )

    # Output journey
    if len(journeys) != 0:
        print_journeys(journeys, dep_secs)


def run_mcraptor(
    timetable: Timetable,
    origin_station: str,
    destination_station: str,
    dep_secs: int,
    rounds: int,
) -> List[Journey]:
    """
    Perform the McRaptor algorithm.

    :param timetable: timetable
    :param origin_station: Name of origin station
    :param destination_station: Name of destation station
    :param dep_secs: Time of departure in seconds
    :param rounds: Number of iterations to perform
    """

    # Get stops for origins and destinations
    from_stops = timetable.stations.get(origin_station).stops
    to_stops = timetable.stations.get(destination_station).stops

    # Run Round-Based Algorithm
    raptor = McRaptorAlgorithm(timetable)
    bag_round_stop = raptor.run(from_stops, dep_secs, rounds)

    # Determine the best destination ID, destination is a platform
    last_round_bag = copy(bag_round_stop[rounds])
    destination_legs = best_legs_to_destination_station(to_stops, last_round_bag)

    if len(destination_legs) != 0:
        logger.debug("Destination leg(s):")
        for leg in destination_legs:
            logger.debug(f"> {leg}")
    else:
        logger.info("Destination unreachable with given parameters")

    journeys = reconstruct_journeys(
        from_stops, destination_legs, bag_round_stop, k=rounds
    )

    return journeys


if __name__ == "__main__":
    args = parse_arguments()
    main(
        args.input,
        args.origin,
        args.destination,
        args.time,
        args.rounds,
    )
